#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import GTK, gtk, GtkExtra
import os, time, re
import string # For backwards compatibility with 1.5.x
import main, app, ftp

class AbstractFileWin:

	UNSORTED = 0
	SORT_BY_NAME = 1
	SORT_BY_SIZE = 2
	SORT_BY_DATE = 3

	def __init__ (self, container, headings):

		# Initialize instance variables
		self.sorted = 1
		self.directories_first = 1
		self.dir = ''

		box = gtk.GtkVBox ()
		box.show ()
		container.add (box)

		self.win = gtk.GtkScrolledWindow ()
		self.win.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		self.win.show ()
		box.pack_start (self.win)

		self.status_bar = gtk.GtkHBox ()
		self.status_bar.set_usize (-1, 15)
		self.status_bar.show ()
		box.pack_start (self.status_bar, expand=gtk.FALSE, fill=gtk.TRUE)

		self.list = gtk.GtkCList (len (headings), headings)
		self.list.connect ("select_row", self.select_items)
		self.list.connect ("button_press_event", self.popup_menu)

		self.list.set_selection_mode (GTK.SELECTION_MULTIPLE)
		self.win.add_with_viewport (self.list)
		self.list.show ()

		self.load_pixmaps ()

	def load_pixmaps (self):

		self.folder_pixmap, self.folder_mask = gtk.create_pixmap_from_xpm (self.list, None, os.path.join (main.ICONS_PREFIX, "folder.xpm"))
		self.link_pixmap, self.link_mask = gtk.create_pixmap_from_xpm (self.list, None, os.path.join (main.ICONS_PREFIX, "link.xpm"))
		self.file_pixmap, self.file_mask = gtk.create_pixmap_from_xpm (self.list, None, os.path.join (main.ICONS_PREFIX, "file.xpm"))

	def select_pixmap (self, listitem):

		file, size, mtime, type, truesize = listitem
		if not type:
			return self.folder_pixmap, self.folder_mask
		else:
			if type[0] == 'l':
				return self.link_pixmap, self.link_mask
			elif type[0] in 'd ':
				return self.folder_pixmap, self.folder_mask
			else:
				return self.file_pixmap, self.file_mask

	def select_items (self, list, r, c, event):

		# This needs to be overridden in the subclass
		pass

	def popup_menu (self, list, event):

		# This needs to be overridden in the subclass
		pass

	def unsorted (self, action, button):

		self.sorted = self.UNSORTED
		self.directories_first = 0
		self.update_listing ('.')

	def sort_by_name (self, action, button):

		self.sorted = self.SORT_BY_NAME
		self.directories_first = 1
		self.update_listing ('.')

	def sort_by_size (self, action, button):

		self.sorted = self.SORT_BY_SIZE
		self.directories_first = 1
		self.update_listing ('.')

	def sort_by_date (self, action, button):

		self.sorted = self.SORT_BY_DATE
		self.directories_first = 1
		self.update_listing ('.')

	def sort_directories_first (self, button):

		self.directories_first = (self.directories_first + 1) % 2
		self.update_listing ('.')

	def clear_items (self):
		self.list.clear ()

	def get_dir (self):
		return self.dir

	def get_file_size (self, file):

		for i in range (self.list.rows):
			if self.list.get_pixtext (i, 0)[0] == file:
				return int (self.list.get_text (i, 4))
		return None

	def update_listing (self, dir):

		# To be overridden in the subclass
		pass

	def sort_list (self, list):

		# To be overridden in the subclass
		pass

	def perform_sort (self, list):

		if self.sorted == self.SORT_BY_NAME:
			list.sort (lambda x, y: cmp (x[0], y[0]))
		elif self.sorted == self.SORT_BY_SIZE:
			list.sort (lambda x, y: cmp (int (x[4]), int (y[4])))
		elif self.sorted == self.SORT_BY_DATE:
			list.sort (lambda x, y: cmp (time.strptime (x[2], "%b %d %H:%M"), time.strptime (y[2], "%b %d %H:%M")))
		return list

class LocalWin (AbstractFileWin):

	headings = [ 'Filename', 'Size', 'Date', 'Type', 'Truesize' ]
	time_re = re.compile ('[A-Za-z]+\s+([A-Za-z]+\s+\d+\s+\d+:\d+):(\d+)\s+\d+')

	UNSORTED = 0
	SORT_BY_NAME = 1
	SORT_BY_SIZE = 2
	SORT_BY_DATE = 3

	def __init__ (self, container):

		AbstractFileWin.__init__ (self, container, self.headings)

		button = gtk.GtkButton ("B")
		button.connect ("clicked", self.enable_dir_update)
		button.set_usize (15, 15)
		button.show ()
		self.status_bar.pack_start (button, expand=gtk.FALSE, fill=gtk.FALSE)

		self.status_label = gtk.GtkLabel ("")
		self.status_label.show ()
		self.status_bar.pack_start (self.status_label, expand=gtk.FALSE, fill=gtk.FALSE, padding=6)

		# Set the settings for our list
		for i in range(len (self.headings)):
			self.list.set_column_width (i, 100)
			self.list.set_column_justification (i, gtk.JUSTIFY_LEFT)
		# Explicitly make the first column for the filename bigger
		self.list.set_column_width (0, 200)
		# Hide the flags column
		self.list.set_column_visibility (3, 0)
		self.list.set_column_visibility (4, 0)

		# Open up the default directory for display
		if os.environ.has_key ('HOME'):
			self.dir = os.environ['HOME']
		else:
			self.dir = os.getcwd ()
		self.update_listing (self.dir)

	def select_items (self, list, r, c, event):

		if hasattr (event, 'type'):
			# Check for double clicks
			if event.type == 5:
				selected = list.get_pixtext (r, 0)[0]

				# Construct our new dir path
				if selected == ".":
					return
				elif selected == "..":
					self.dir = os.path.dirname (self.dir)
				else:
					newpath = os.path.join (self.dir, selected)
					if not os.path.isdir (newpath):
						return
					else:
						self.dir = newpath

				# If we've made it this far, load the new dir
				self.update_listing (self.dir)

	def popup_menu (self, list, event):

		if event.button == 3:
			opts = [
				('/Upload Files',             None, self.upload_selected,  0, ''),
				#('/Open',                     None, self.open_selected,    0, ''),
				#('/Edit',                     None, self.edit_selected,    0, ''),
				#('/View File',                None, None,                  0, ''),
				('/Delete Selected Files',    None, self.delete_selected,  0, ''),
				('/Rename File',              None, self.rename_selected,  0, ''),
				('/Create Directory',         None, self.create_directory, 0, ''),
				#('/Change Directory',         None, None,                  0, ''),
				#('/Sep1',                     None, None,                  0, '<Separator>'),
				#('/Change Drive',             None, None,                  0, ''),
				#('/Sep2',                     None, None,                  0, '<Separator>'),
				#('/Compare Directories',      None, None,                  0, ''),
				#('/Open Current in Browser',  None, None,                  0, ''),
				#('/Open Selected in Browser', None, None,                  0, ''),
				#('/Command Prompt Here',      None, None,                  0, ''),
				#('/Sep3',                     None, None,                  0, '<Separator>'),
				#('/Directory Info',           None, None,                  0, ''),
				('/Sep4',                     None, None,                  0, '<Separator>'),
				('/Sort Order',               None, None,                  0, '<Branch>'),
				('/Sort Order/Unsorted',      None, self.unsorted,         0, ''),
				('/Sort Order/Sort by Name',  None, self.sort_by_name,     0, ''),
				('/Sort Order/Sort by Size',  None, self.sort_by_size,     0, ''),
				('/Sort Order/Sort by Date',  None, self.sort_by_date,     0, ''),
				('/Sort Order/<separator>',   None, None,                  0, '<Separator>'),
				('/Sort Order/Order Directories-Links-Files', None, None, 0, '<ToggleItem>')
			]

			ag = gtk.GtkAccelGroup()
			items = gtk.GtkItemFactory(gtk.GtkMenu.get_type(), "<popup>", ag)
			items.create_items (opts)

			checked = items.get_widget ('/Sort Order/Order Directories-Links-Files')
			if self.directories_first:
				checked.set_active (gtk.TRUE)
			checked.connect ("activate", self.sort_directories_first)

			menu = items.get_widget ("<popup>")
			menu.popup (None, None, None, event.button, event.time)
			menu.show ()

	def upload_selected (self, action, button):

		files = [ ]

		for i in self.list.selection:
			entry = [ ]
			entry.append (self.list.get_pixtext (i, 0)[0])
			# Get remaining entries
			for j in range (self.list.columns - 1):
				entry.append (self.list.get_text (i, j + 1))

			if entry[0] == "..":
				continue
			else:
				files.append (tuple (entry))

		self.list.unselect_all ()
		main.app.transfer (main.app.get_remote_dir (), self.get_dir (), files, ftp.UPLOAD)

	def open_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			path = os.path.join (self.get_dir (), file)

			# Skip the home directory entry
			if file == ".." or \
			   not os.path.isfile (path) or \
			   not os.access (path, os.X_OK):
				continue
			else:
				# Execing code will go here
				pass

	def edit_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			path = os.path.join (self.get_dir (), file)

			# Editing code will go here

	def delete_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			type = self.list.get_text (i, 3)[0]
			path = os.path.join (self.dir, file)
			if type in '-lf':
				os.remove (path)
			else:
				os.rmdir (path)
		self.list.unselect_all ()
		self.update_listing (self.dir)

	def rename_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			orig_path = os.path.join (self.get_dir (), file)
			new = GtkExtra.input_box (title="FtpCube - Rename", message="Rename %s to:" %file)
			if new:
				new_path = os.path.join (self.get_dir (), new)
				try:
					os.rename (orig_path, new_path)
				except OSError, strerror:
					GtkExtra.message_box (title="FtpCube Error", message="FtpCube error: %s" %strerror, buttons=( [ "Ok" ]))
					return
		self.update_listing (self.dir)

	def create_directory (self, action, button):

		dir = GtkExtra.input_box (title="FtpCube - Create Directory", message="Directory Name:")
		path = os.path.join (self.get_dir (), dir)
		try:
			os.mkdir (path)
		except OSError, strerror:
			GtkExtra.message_box (title="FtpCube Error", message="FtpCube error: %s" %strerror, buttons=( [ "Ok" ]))
			return
		self.update_listing (self.dir)

	def update_status_dir (self, dir):
		self.status_label.set_text (dir)

	def enable_dir_update (self, button):
		pass

	def read_dir (self, dir):

		# If we say '.', we really mean our current directory
		if dir == '.':
			dir = self.dir

		try:
			files = os.listdir (dir)
		except OSError, (errno, strerror):
			GtkExtra.message_box (title="FtpCube Error", message="FtpCube Error: %s" %strerror, buttons=( [ "Ok" ]))
			return

		# Add the special file ".."
		files.insert (0, "..")

		# Now grab the file info and load it into the load
		listitems = []
		for f in files:
			(st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size,
			 st_atime, st_mtime, st_ctime) = os.stat (os.path.join (dir, f))

			# Beautify the size string
			size = app.beautify_size (st_size)

			# Beautify the time
			match = self.time_re.match (time.ctime (st_mtime))
			mtime = match.group (1)

			# Set the appropriate file type
			if os.path.islink (os.path.join (self.dir, f)):
				type = "l"
			elif os.path.isdir (os.path.join (self.dir, f)):
				type = "d"
			else:
				type = "f"

			listitems.append ((f, size, mtime, type, "%s" %st_size))

		return listitems

	def update_listing (self, dir):

		listitems = self.read_dir (dir)

		# Now change the list to reflect the new listing. Notice that we
		# pause the visuals during the update so it doesn't look funny
		self.list.freeze ()
		self.clear_items ()
		if self.sorted:
			listitems = self.sort_list (listitems)
		for i in range (len (listitems)):
			self.list.append (listitems[i])
			listitem = listitems[i]
			pixmap, mask = self.select_pixmap (listitem)
			self.list.set_pixmap (i, 0, pixmap, mask)
			self.list.set_pixtext (i, 0, listitem[0], 4, pixmap, mask)
		self.list.thaw ()

		# Update the status
		self.update_status_dir (dir)

	def sort_list (self, list):

		if self.directories_first:
			dirs = filter (lambda i: i[3] == 'd', list)
			links = filter (lambda i: i[3] == 'l', list)
			files = filter (lambda i: i[3] in 'f-', list)

			list = [ ]
			list.extend (self.perform_sort (dirs))
			list.extend (self.perform_sort (links))
			list.extend (self.perform_sort (files))
		else:
			list = self.perform_sort (list)
		return list

class RemoteWin (AbstractFileWin):

	headings = [ 'Filename', 'Size', 'Date', 'Flags', 'Truesize' ]

	list_re = re.compile ('([A-Za-z-]+)\s+(\d+)\s+([A-Za-z0-9]+)\s+([A-Za-z0-9]+)\s+(\d+)\s+([A-Za-z]+\s+[0-9]+\s+[0-9:]+)\s+(.*)', re.IGNORECASE)
	link_re = re.compile ('(.*) -> (.*)')
	broken_link_re = re.compile ('.*-> (.*)')
	time_re = re.compile ('[A-Za-z]+\s+([A-Za-z]+\s+\d+\s+\d+:\d+):(\d+)\s+\d+')

	UNSORTED = 0
	SORT_BY_NAME = 1
	SORT_BY_SIZE = 2
	SORT_BY_DATE = 3

	def __init__ (self, container):

		AbstractFileWin.__init__ (self, container, self.headings)

		self.status_label = gtk.GtkLabel ()
		self.status_label.show ()
		self.status_bar.pack_start (self.status_label, expand=GTK.FALSE, padding=6)

		# Set the list settings for our file window
		for i in range(len (self.headings)):
			self.list.set_column_width (i, 100)
			self.list.set_column_justification (i, gtk.JUSTIFY_LEFT)
		# Explicitly make the first column for the filename bigger
		self.list.set_column_width (0, 200)
		self.list.set_column_visibility (4, 0)

	def select_items (self, list, r, c, event):

		if hasattr (event, 'type'):
			# Check for double clicks
			if event.type == 5:
				selected = list.get_pixtext (r, 0)[0]

				# Construct our new dir path
				if selected == "..":
					main.app.main_thread.cwd ('..')
					main.app.main_thread.list ()

					# Update the status directory
					text = self.get_dir ()
					index = string.rfind (text, "/")
					if index == 0:
						self.update_status_dir ("/")
					else:
						self.update_status_dir (text[:index])
				else:
					# Check if we have a directory
					f = list.get_text (r, 3)[0]
					if f in 'ld':

						status_text = self.get_dir ()

						def exception_callback (errmsg, status_text=status_text, self=self):

							gtk.threads_enter ()
							main.app.ftp_error_tell_user (errmsg)
							self.update_status_dir (status_text)
							gtk.threads_leave ()

						# Update the status directory before the action
						# so we can roll back in the case of an exception
						if status_text == "/":
							self.update_status_dir ("/" + selected)
						else:
							self.update_status_dir ("%s/%s" %(status_text, selected))

						main.app.main_thread.set_exception_handler (exception_callback)
						main.app.main_thread.cwd (selected)
						main.app.main_thread.set_exception_handler (exception_callback)
						main.app.main_thread.list ()

	def popup_menu (self, list, event):

		if event.button == 3:
			opts = [
				("/Download Selected",         None, self.download_selected, 0, ''),
				#("/View File",                 None, None, 0, ''),
				("/Delete Selected",           None, self.delete_selected, 0, ''),
				#("/Recursive Delete",          None, None, 0, ''),
				#("/Rename",                    None, None, 0, ''),
				#("/Create Directory",          None, None, 0, ''),
				#("/Sep1",                      None, None, 0, '<Separator>'),
				#("/Change Directory",          None, None, 0, ''),
				#("/Sep2",                      None, None, 0, '<Separator>'),
				#("/Compare Directories",       None, None, 0, ''),
				#("/Add Bookmarks",             None, None, 0, ''),
				#("/Sep3",                      None, None, 0, '<Separator>'),
				#("/Set Attributes",            None, None, 0, ''),
				#("/Custom Commands",           None, None, 0, ''),
				#("/Directory Info",            None, None, 0, ''),
				("/Sep4",                      None, None, 0, '<Separator>'),
				("/Sort Order",                None, None, 0, '<Branch>'),
				("/Sort Order/Unsorted",       None, self.unsorted, 0, ''),
				("/Sort Order/Sort by Name",   None, self.sort_by_name, 0, ''),
				("/Sort Order/Sort by Size",   None, self.sort_by_size, 0, ''),
				("/Sort Order/Sort by Date",   None, self.sort_by_date, 0, ''),
				("/Sort Order/Sep5",           None, None, 0, '<Separator>'),
				("/Sort Order/Order Directories-Links-Files", None, None, 0, '<ToggleItem>')
			]

			ag = gtk.GtkAccelGroup()
			items = gtk.GtkItemFactory(gtk.GtkMenu.get_type(), "<popup>", ag)
			items.create_items (opts)

			checked = items.get_widget ('/Sort Order/Order Directories-Links-Files')
			if self.directories_first:
				checked.set_active (gtk.TRUE)
			checked.connect ("activate", self.sort_directories_first)

			menu = items.get_widget ("<popup>")
			menu.popup (None, None, None, event.button, event.time)
			menu.show ()

	def update_listing (self, listing):

		# Create the ".." entry
		match = self.time_re.match (time.ctime (time.time ()))
		localtime = match.group (1)
		dotdotentry = ('..', '4.00 KBytes', localtime, '', '0')
		listitems = [ dotdotentry ]

		# Update the last listing array and parse the entries to update
		# the remote listing
		for entry in listing:
			tuple = self.parse_list_entry (entry)
			if tuple is None:
				continue
			else:
				file, truesize, date, flags = tuple

			# Beautify the size string
			size = app.beautify_size (truesize)

			# Check if the file is a link and if so, clean up the file name
			if flags and flags[0] == 'l':
				match = self.link_re.match (file)
				if match is not None:
					file, dest = match.groups ()
				else:
					# Try the broken link match
					match = self.broken_link_re.match (file)
					if match is not None:
						file = match.group (1)
						file = os.path.basename (file)
					else:
						print "WARNING: Match failed for: %s" %file

			listitems.append ((file, size, date, flags, "%s" %truesize))

		# Now change the list to reflect the new listing. Notice that we
		# pause the visuals during the update so it doesn't look funny
		if self.sorted:
			listitems = self.sort_list (listitems)
		self.set_list (listitems)

	def get_status (self):
		return self.status_label.get ()

	def update_status_dir (self, dir):

		self.dir = dir
		self.update_status (self.dir)

	def update_status (self, text):
		self.status_label.set_text (text)

	def parse_list_entry (self, entry):

		match = self.list_re.match (entry)
		if match is None:
			#print "WARNING: listing match failed for: %s" %entry
			return None
		else:
			flags, hardlinks, user, group, size, date, file =  match.groups ()
			size = int (size)
			return file, size, date, flags

	def download_selected (self, action, button):

		files = [ ]

		# Extract the data from the CList. Note we don't use
		# get_row_data because it appears to be broken
		for i in self.list.selection:
			entry = [ ]
			entry.append (self.list.get_pixtext (i, 0)[0])
			# Get rest of entries
			for j in range (self.list.columns - 1):
				entry.append (self.list.get_text (i, j + 1))

			# Discard the ".." entry
			if entry[0] == "..":
				continue

			# Fix up the type
			if not entry[3][0] == '-':
				entry[3] = entry [3][0]
			else:
				entry[3] = 'f'

			files.append (tuple (entry))

		self.list.unselect_all ()
		main.app.transfer (self.get_dir (), main.app.get_local_dir (), files, ftp.DOWNLOAD)

	def delete_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			type = self.list.get_text (i, 3)[0]
			if type in 'l-':
				main.app.main_thread.delete (file)
			else:
				main.app.main_thread.rmdir (file)

		self.list.unselect_all ()
		main.app.main_thread.list ()

	def set_list (self, list):

		self.list.freeze ()
		self.clear_items ()
		for i in range (len (list)):
			self.list.append (list[i])
			listitem = list[i]
			pixmap, mask = self.select_pixmap (listitem)
			self.list.set_pixmap (i, 0, pixmap, mask)
			self.list.set_pixtext (i, 0, listitem[0], 4, pixmap, mask)
		self.list.thaw ()

	def get_list (self):

		list = [ ]
		for i in range (self.list.rows):
			entry = [ ]
			for j in range (self.list.columns):
				if j == 0:
					entry.append (self.list.get_pixtext (i, j)[0])
				else:
					entry.append (self.list.get_text (i, j))
			list.append (tuple (entry))
		return list

	def sort_list (self, list):

		if self.directories_first:
			dirs = filter (lambda i: not i[3] or i[3][0] == 'd', list)
			links = filter (lambda i: i[3] and i[3][0] == 'l', list)
			files = filter (lambda i: i[3] and i[3][0] in 'f-', list)

			list = [ ]
			list.extend (self.perform_sort (dirs))
			list.extend (self.perform_sort (links))
			list.extend (self.perform_sort (files))
		else:
			list = self.perform_sort (list)
		return list
