#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk, GtkExtra
import string # For backwards compatibility with 1.5.x
import socket, threading, re, os, errno, time
import exceptions
import main, app, logging

CRLF = '\r\n'

DOWNLOAD = 0
UPLOAD = 1

class FtpException (exceptions.Exception):

	def __init__ (self, args=None):
		self.args = args

class FtpControlConnection (threading.Thread):

	def __init__ (self, host, port, silence=0, logging=0):

		threading.Thread.__init__ (self)
		self.setDaemon (1)

		# Initialize instance variables
		self.sock = None
		self.file = None
		self.multiline_code = None
		self.cmdobj = None
		self.exception_handler = None
		self.incoming = [ ]
		self.outgoing = [ ]
		self.outgoing_event = threading.Event ()
		self.die = 0
		self.host = host
		self.port = port
		self.silence = silence
		self.cmd_log = None
		self.set_log_active (logging)

	def connect (self):

		# Create the socket
		proto = socket.getprotobyname ('tcp')
		try:
			self.sock = socket.socket (socket.AF_INET, socket.SOCK_STREAM, proto)
		except socket.error, strerror:
			raise FtpException ("Error opening control socket: %s" %strerror)

		gtk.threads_enter ()
		if not self.silence: main.app.ftp_local_tell_user ("Connecting to %s on port %d." %(self.host, self.port))
		gtk.threads_leave ()

		self.sock.setblocking (0)

		try:
			err = self.sock.connect_ex ((self.host, self.port))
		except socket.error, strerror:
			# If we're in here, we probably caught an error such as hostname lookups
			# This is pretty fatal, so let's clear any exception handlers before raising
			# the exception
			self.clear_exception_handler ()
			raise FtpException ("Error connecting to host %s: %s" %(self.host, strerror))

		gtk.threads_enter ()
		if not self.silence: main.app.ftp_local_tell_user ("Connected to %s." %self.host)
		gtk.threads_leave ()

		while err == errno.EINPROGRESS and not self.die:

			time.sleep (1)
			err = self.sock.getsockopt (socket.SOL_SOCKET, socket.SO_ERROR)

		if err != 0 or self.die:
			raise FtpException ("Error connecting to host %s: %s" %(self.host, os.strerror (err)))
		else:
			self.sock.setblocking (1)
			self.file = self.sock.makefile ('rb')

	def destroy (self):

		# This might need some revising at some point
		if not self.silence: main.app.ftp_local_tell_user ("Closing connection to %s..." %(self.host))
		self.die = 1
		self.outgoing_event.set ()
		if self.file is not None:
			self.file.close ()
		if self.sock is not None:
			self.sock.close ()
		if self.logging:
			self.log ("Closing connection to %s:%d..." %(self.host, self.port))
			self.log ('')

	def get_host (self):
		return self.host

	def get_port (self):
		return self.port

	def reset_queue (self):

		self.outgoing = [ ]
		self.outgoing_event.set ()

	def wait_for_response (self):

		while 1:
			line = self.file.readline ()

			if not line:
				gtk.threads_enter ()
				if not self.silence: main.app.ftp_local_tell_user ("Connection to %s terminated..." %self.host)
				gtk.threads_leave ()
				self.sock.close ()
				return

			# Clean up the line
			if line[-2:] == CRLF: line = line[:-2]
			elif line[-1:] in CRLF: line = line[:-1]

			data = self.process_data (line)
			if data is not None:
				return data

	def process_data (self, line):

		if not self.silence:
			gtk.threads_enter ()
			main.app.ftp_remote_tell_user (line)
			gtk.threads_leave ()

		if self.logging:
			self.log (line)

		self.incoming.append (line)

		if line[3:4] == '-' or self.multiline_code is not None:
			if self.multiline_code is None:
				self.multiline_code = line[:3]
				return
			elif line[:3] == self.multiline_code and line[3:4] != '-':
				self.multiline_code = None
				data = self.incoming
				self.incoming = [ ]
				return data
		else:
			data = self.incoming
			self.incoming = [ ]
			return data

	def send (self, msg):

		if self.logging:
			self.log (msg)
		return self.sock.send (msg + CRLF)

	def send_oob (self, msg):
		return self.sock.send (msg + CRLF, socket.MSG_OOB)

	def run (self):

		while not self.die:
			if self.outgoing == [ ]:
				self.outgoing_event.clear ()
				self.outgoing_event.wait ()

			if not self.outgoing == [ ]:
				self.cmdobj = self.outgoing.pop (0)
				if self.cmdobj is None:
					continue
				else:
					try:
						# Call the execute function directly so
						# it runs in our thread
						self.cmdobj.execute ()
					except Exception, strerror:

						# Always reset the queue
						self.reset_queue ()

						if not self.die:
							if self.exception_handler is not None:
								self.exception_handler (strerror)
							else:
								gtk.threads_enter ()
								main.app.ftp_error_tell_user (strerror)
								gtk.threads_leave ()

					# Always clear handler after execution
					self.clear_exception_handler ()

	def set_exception_handler (self, callback):
		self.exception_handler = callback

	def clear_exception_handler (self):
		self.exception_handler = None

	def get_executing_obj (self):
		return self.cmdobj

	def busy (self):
		return self.outgoing_event.isSet ()

	def add_outgoing (self, cmdobj):

		self.outgoing.append (cmdobj)
		if not self.outgoing_event.isSet ():
			self.outgoing_event.set ()

	def set_log_active (self, bool):

		if bool:
			if self.cmd_log is None:
				self.cmd_log = logging.Log (os.path.join (app.CONFIG_DIR, main.app['cmd_log']))
		else:
			if self.cmd_log is not None:
				del self.cmd_log
				self.cmd_log = None
		self.logging = bool

	def log (self, cmd):
		self.cmd_log.write (cmd)

class FTPGetWelcomeMsg:

	def __init__ (self, queue):
		self.control_queue = queue

	def execute (self):

		response = self.control_queue.wait_for_response ()
		c = response[-1][:1]
		if c not in '123':
			raise FtpException ("Error connecting to server: Welcome message not received correctly")

class FTPConnect:

	def __init__ (self, queue):
		self.control_queue = queue

	def execute (self):
		self.control_queue.connect ()

class FTPAbortCmd:

	def __init__ (self, queue, silence=0):

		self.control_queue = queue
		self.silence = silence

	def execute (self):

		if not self.silence:
			gtk.threads_enter ()
			main.app.ftp_local_tell_user ("ABOR")
			gtk.threads_leave ()

		self.control_queue.send_oob ("ABOR")
		response = self.control_queue.wait_for_response ()
		if response[-1][:3] not in ('426', '226'):
			raise FtpException (response[-1])

class FTPVoidCmd:

	def __init__ (self, queue, cmd, silence=0):

		self.control_queue = queue
		self.silence = silence
		self.cmd = cmd

	def execute (self):

		if not self.silence:
			gtk.threads_enter ()
			main.app.ftp_local_tell_user (self.cmd)
			gtk.threads_leave ()

		self.control_queue.send (self.cmd)
		self.response = self.control_queue.wait_for_response ()
		if self.response[-1][0] != '2':
			raise FtpException (self.response[-1])

class FTPBasicCmd (FTPVoidCmd):

	def __init__ (self, queue, cmd, silence=0):
		FTPVoidCmd.__init__ (self, queue, cmd, silence)

	def get_response (self):
		return self.response

class FTPLogin:

	def __init__ (self, queue, user, pw, acct, silence=0):

		self.control_queue = queue
		self.silence = silence
		self.user = user
		self.pw = pw
		self.acct = acct

	def execute (self):

		string = "USER " + self.user

		if not self.silence:
			gtk.threads_enter ()
			main.app.ftp_local_tell_user (string)
			gtk.threads_leave ()

		self.control_queue.send (string)
		response = self.control_queue.wait_for_response ()

		if response[-1][0] == '3':

			string = "PASS " + self.pw

			if not self.silence:
				gtk.threads_enter ()
				main.app.ftp_local_tell_user (string)
				gtk.threads_leave ()

			self.control_queue.send (string)
			response = self.control_queue.wait_for_response ()

		if response[-1][0] == '3':
			# For now, assume ACCT to be ''
			string = "ACCT "

			if not self.silence:
				gtk.threads_enter ()
				main.app.ftp_local_tell_user (string)
				gtk.threads_leave ()

			self.control_queue.send (string)
			response = self.control_queue.wait_for_response ()

		if response[-1][0] != '2':
			raise FtpException ("Error logging in to %s" %self.control_queue.get_host ())

class FTPTransferConnection:

	def __init__ (self, control, host, port, passive=1):

		self.control_queue = control
		self.abort_flag = 0

		if passive:
			# Create the socket
			proto = socket.getprotobyname ('tcp')
			try:
				self.sock = socket.socket (socket.AF_INET, socket.SOCK_STREAM, proto)
			except socket.error, strerror:
				raise FtpException ("Error creating socket for transfer: %s" %strerror)

			self.sock.setblocking (0)

			try:
				err = self.sock.connect_ex ((host, port))
			except socket.error, strerror:
				raise FtpException ("Error creating socket for transfer: %s" %(strerror))

			while err == errno.EINPROGRESS and not self.abort_flag:

				time.sleep (1)
				err = self.sock.getsockopt (socket.SOL_SOCKET, socket.SO_ERROR)

			if err != 0 or self.abort_flag:
				raise FtpException ("Error creating socket for transfer: %s" %os.strerror (err))
			else:
				self.sock.setblocking (1)
				self.file = self.sock.makefile ('rb')
		else:
			host, port = self.create_server_sock ()
			try:
				self.sock = socket.socket (socket.AF_INET, socket.SOCK_STREAM)
				self.sock.bind (('', 0))
				self.sock.listen (1)
				dummyhost, port = self.sock.getsockname ()
				host, dummyport = control.getsockname ()
			except socket.error, (strerror):
				raise FtpException ("Error creating socket for transfer: %s" %strerror)

			self.send_port (host, port)
			self.sock, sockaddr = self.sock.accept ()

	def send_port (self, host, port):

		hbytes = string.split (host, '.')
		pbytes = [ `port/256`, `port%256` ]
		bytes = hbytes + pbytes
		cmd = 'PORT ' + string.join (bytes, ',')
		return self.control.basic_cmd (cmd)

	def create_server_sock (self):
		pass

	def get_sock (self):
		return self.sock

	def get_file (self):
		return self.sock.makefile ('rb')

	def close (self):
		self.sock.close ()

	def abort (self):
		self.abort_flag = 1

class FTPTransferCmd:

	# For parsing server return codes
	re150 = re.compile ("150 .* \((\d+) bytes\)", re.IGNORECASE)

	def __init__ (self, queue, cmd, silence=0):

		self.control_queue = queue
		self.silence = silence
		self.cmd = cmd

	def execute (self):

		self.control_queue.send (self.cmd)
		response = self.control_queue.wait_for_response ()

		if response[-1][0] != '1':
			raise FtpException ("Error initiating transfer: %s" %response[-1])
		if response[-1][:3] == '150':
			self.size = self.parse150 (response[-1])

	def parse150 (self, resp):

		# Sanity checking
		if resp[:3] != '150':
			raise FtpException ("Error parsing 150 response: 150 response not given")

		matched = self.re150.match (resp)
		if matched:
			return int (matched.group (1))
		return None

	def get_size (self):
		return self.size


class FTPTransfer:

	def __init__ (self, queue, rest=None, passive=1, silence=0):

		self.control_queue = queue
		self.silence = silence
		self.rest = rest
		self.passive = passive

	def initiate_transfer (self):

		if self.passive:
			cmd = FTPBasicCmd (self.control_queue, 'PASV', silence=self.silence)
			cmd.execute ()
			host, port = self.parse227 ( cmd.get_response()[-1] )

			self.transfer = FTPTransferConnection (self.control_queue, host, port, passive=self.passive)

			if self.rest is not None:
				cmd = FTPVoidCmd (self.control_queue, "REST %s" %self.rest, silence=self.silence)
				cmd.execute ()

			if not self.silence:
				gtk.threads_enter ()
				main.app.ftp_local_tell_user (self.cmd)
				gtk.threads_leave ()

			cmd = FTPTransferCmd (self.control_queue, self.cmd, silence=self.silence)
			cmd.execute ()

			self.size = cmd.get_size ()
		else:
			# Non passive stuff will go here
			pass

	def parse227 (self, resp):

		# Sanity checking
		if resp[:3] != '227':
			raise FtpException ("Error parsing 227 response: 227 response not given")

		left_bracket = string.find (resp, '(')
		right_bracket = string.find (resp, ')')
		if left_bracket < 0 or right_bracket < 0:
			raise FtpException ("Error parsing 227 response: Format should be (h1,h2,h3,h4,p1,p2) and given: %s" %resp)
		fields = string.split (resp[left_bracket + 1: right_bracket], ',')
		if len (fields) != 6:
			raise FtpException ("Error parsing 227 response: Unable to pack appropriate number of fields")
		host = string.join (fields[:4], '.')
		port = (int (fields[4]) << 8) + int (fields[5])
		return host, port

	def abort (self):
		self.transfer.abort ()

class FTPList (FTPTransfer):

	def __init__ (self, queue, cmd, callback=None, rest=None, passive=1, silence=0):

		FTPTransfer.__init__ (self, queue, rest=rest, passive=passive, silence=silence)
		self.control_queue = queue
		self.silence = silence
		self.callback = callback
		self.cmd = cmd
		self.abort_flag = 0
		self.data = [ ]              # Local copy of data

	def execute (self):

		# Set the appropriate transfer type
		type = FTPBasicCmd (self.control_queue, 'TYPE A', silence=self.silence)
		type.execute ()

		# Now prepare for the transfer
		self.initiate_transfer ()

		self.sock = self.transfer.get_sock ()
		self.file = self.transfer.get_file ()

		bytes = 0

		while not self.abort_flag:
			line = self.file.readline ()
			if not line:
				break
			if line[-2:] == CRLF:
				line = line[:-2]
			elif line[-1:] in CRLF:
				line = line[:-1]

			# For updating the status window
			bytes = bytes + len (line)
			gtk.threads_enter ()
			main.app.update_list_transfer (bytes)
			gtk.threads_leave ()

			self.data.append (line)

		# Wait for the finished transfer and then get the response
		self.file.close ()
		self.sock.close ()
		self.control_queue.wait_for_response ()

		# Now reset the status window to the current directory
		gtk.threads_enter ()
		main.app.list_transfer_finished ()
		gtk.threads_leave ()

		if self.callback is None:
			gtk.threads_enter ()
			main.app.update_remote_listing (self.get_data ())
			gtk.threads_leave ()
		else:
			self.callback (self.get_data ())

	def get_data (self):
		return self.data

	def abort (self):

		self.transfer.abort ()
		self.abort_flag = 1

class FTPBinaryTransfer (FTPTransfer):

	def __init__ (self, queue, cmd, local_file, direction, progress=None, blocksize=8192, rest=None, passive=1, silence=0):

		FTPTransfer.__init__ (self, queue, rest, passive)
		self.control_queue = queue
		self.cmd = cmd
		self.local_file = local_file
		self.direction = direction
		self.silence = silence
		self.progress = progress
		self.blocksize = blocksize
		self.rest = rest
		self.passive = passive
		self.abort_flag = 0               # For aborting the transfer

	def execute (self):

		# Set the appropriate transfer type
		type = FTPBasicCmd (self.control_queue, 'TYPE I', silence=self.silence)
		type.execute ()

		# Now prepare for the transfer
		self.initiate_transfer ()

		self.sock = self.transfer.get_sock ()

		if self.direction == DOWNLOAD:
			file = open (self.local_file, 'w')
			while not self.abort_flag:
				data = self.sock.recv (self.blocksize)
				if not data:
					break
				if self.progress is not None:
					gtk.threads_enter ()
					self.progress.update_transfer (len (data))
					gtk.threads_leave ()
				file.write (data)
		elif self.direction == UPLOAD:
			file = open (self.local_file, 'r')
			while not self.abort_flag:
				data = file.read (self.blocksize)
				if not data:
					break
				self.sock.send (data)
				gtk.threads_enter ()
				self.progress.update_transfer (len (data))
				gtk.threads_leave ()

		file.close ()
		self.sock.close ()

		if self.direction == DOWNLOAD:
			# Wait for the finished transfer and then get the response
			self.control_queue.wait_for_response ()
			gtk.threads_enter ()
			main.app.add_to_download_list (self.local_file)
			main.app.update_local_listing ('.')
			gtk.threads_leave ()
		elif self.direction == UPLOAD:
			self.control_queue.wait_for_response ()
			gtk.threads_enter ()
			main.app.main_thread.list ()
			gtk.threads_leave ()

	def abort (self):
		self.abort_flag = 1

class FTP:

	def __init__ (self, host=None, port=None, silence=0, logging=0):

		# Initialize instance variables
		self.passive_server = 1
		self.silence = silence

		# Sanity checking on our options
		if host is None:
			raise FtpException ("Error opening FTP connection: No host given")
		if port is None or not (port > 0 and port < 65534):
			raise FtpException ("Error opening FTP connection: Invalid port given")

		self.control_connect = FtpControlConnection (host, port, silence=self.silence, logging=logging)
		self.control_connect.start ()

	def destroy (self):
		self.control_connect.destroy ()

	def set_exception_handler (self, callback):
		self.control_connect.set_exception_handler (callback)

	def clear_exception_handler (self):
		self.control_connect.clear_exception_handler ()

	def set_log_active (self, bool):
		self.control_connect.set_log_active (bool)

	def reset (self):
		self.control_connect.reset_queue ()

	def ftp_busy (self):
		return self.control_connect.busy ()

	def initiate_connect (self):

		connect = FTPConnect (self.control_connect)
		self.control_connect.add_outgoing (connect)

	def get_welcome_msg (self):

		welcome = FTPGetWelcomeMsg (self.control_connect)
		self.control_connect.add_outgoing (welcome)

	def perform_login (self, username, password):

		login = FTPLogin (self.control_connect, username, password, '', silence=self.silence)
		self.control_connect.add_outgoing (login)

	def noop (self):

		cmd = 'NOOP'
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.control_connect.add_outgoing (ftpcmd)

	def cwd (self, dir):

		# Default to '/' if dir is blank
		if not dir:
			dir = '/'

		if dir == '..':
			cmd = "CDUP"
		elif dir == ".":
			# Actually do nothing and return
			return
		else:
			cmd = "CWD " + dir
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.control_connect.add_outgoing (ftpcmd)

	def delete (self, file):

		cmd = 'DELE ' + file
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.control_connect.add_outgoing (ftpcmd)

	def mkdir (self, directory):

		cmd = 'MKD ' + directory
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.control_connect.add_outgoing (ftpcmd)

	def rmdir (self, directory):

		cmd = 'RMD ' + directory
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.control_connect.add_outgoing (ftpcmd)

	def abort (self):

		self.control_connect.reset_queue ()
		aborcmd = FTPAbortCmd (self.control_connect, silence=self.silence)
		self.control_connect.add_outgoing (aborcmd)

		# Check if we're cancelling an outstanding transfer such as a binary
		# transfer or list
		executing = self.control_connect.get_executing_obj ()
		if isinstance (executing, FTPTransfer):
			executing.abort ()
		elif isinstance (executing, FTPConnect):
			self.control_connect.destroy ()

	def list (self, args=None, callback=None, rest=None):

		cmd = 'LIST'
		if args is not None:
			for i in args:
				if i:
					cmd = cmd + (' ' + i)
		listcmd = FTPList (self.control_connect, cmd, callback=callback, silence=self.silence)
		self.control_connect.add_outgoing (listcmd)

	def retrieve_binary (self, file, local_dir, progress, blocksize=8192, passive=1, rest=None):

		cmd = 'RETR ' + file
		local_file = os.path.join (local_dir, file)
		binarycmd = FTPBinaryTransfer (self.control_connect, cmd, local_file, DOWNLOAD, progress, blocksize, rest, passive, silence=self.silence)
		self.control_connect.add_outgoing (binarycmd)

	def upload_binary (self, file, local_dir, progress, blocksize=8192, passive=1, rest=None):

		cmd = 'STOR ' + file
		local_file = os.path.join (local_dir, file)
		binarycmd = FTPBinaryTransfer (self.control_connect, cmd, local_file, UPLOAD, progress, blocksize, rest, passive, silence=self.silence)
		self.control_connect.add_outgoing (binarycmd)
